module CustomFielder
  class Field < ActiveRecord::Base
    include TypeCheckingHelper
    include DeserializationHelper

    FIELD_TYPE_MAP = {
      'Text' => 'String',
      'Integer' => 'Fixnum',
      'Decmial' => 'Float',
      'Date' => 'Date',
      'Date & Time' => 'DateTime',
      'True/False' => 'Boolean',
      'List' => 'Array'
    }

    FORM_FIELD_TYPES = {
      'Text - Entry' => :text_entry_fields,
      'Text - Multiple Choice' => :text_choice_fields,
      'Integer - Entry' => :int_entry_fields,
      'Integer - Multiple Choice' => :int_choice_fields,
      'Decimal - Entry' => :float_entry_fields,
      'Decimal - Multiple Choice' => :float_choice_fields,
      'Date - Entry' => :date_entry_fields,
      'Date - Multiple Choice' => :date_choice_fields,
      'Date & Time - Entry' => :datetime_entry_fields,
      'Date & Time - Multiple Choice' => :datetime_choice_fields,
      'True/False' => :boolean_fields
    }

    validate  :default_type
    validate  :field_options_types
    validate  :field_options_unique

    validates :name,                presence: true,
                                    uniqueness: {
                                      scope: [:fieldable_type],
                                      message: "Field name already exists"
                                    }
    validates :field_type,          presence: true,
                                    inclusion: { in: %w[Array Fixnum Float Date DateTime Boolean String] }
    validates :fieldable_type,      presence: true




    has_many :values, dependent: :destroy

    scope :text_entry_fields,       -> { all.where(field_type: 'String', field_options: nil) }
    scope :text_choice_fields,      -> { all.where(field_type: 'String').where.not(field_options: nil) }
    scope :int_entry_fields,        -> { all.where(field_type: 'Fixnum', field_options: nil) }
    scope :int_choice_fields,       -> { all.where(field_type: 'Fixnum').where.not(field_options: nil) }
    scope :float_entry_fields,      -> { all.where(field_type: 'Float', field_options: nil) }
    scope :float_choice_fields,     -> { all.where(field_type: 'Float').where.not(field_options: nil) }
    scope :date_entry_fields,       -> { all.where(field_type: 'Date', field_options: nil) }
    scope :date_choice_fields,      -> { all.where(field_type: 'Date').where.not(field_options:nil) }
    scope :datetime_entry_fields,   -> { all.where(field_type: 'DateTime', field_options: nil) }
    scope :datetime_choice_fields,  -> { all.where(field_type: 'DateTime').where.not(field_options:nil) }
    scope :boolean_fields,          -> { all.where(field_type: 'Boolean') }

    ##
    # Creates a hash of form field types mapped to array of field, field id pairs
    # meant for use in a form select dropdown with options group
    #
    # @return [Hash]
    #
    def self.all_fields_hashed_for_opt_group
      FORM_FIELD_TYPES.inject(Hash.new) do |hash, form_field|
        fields = send(form_field.last)
        hash[form_field.first] = fields.map { |field| [field.name, field.id] } unless fields.empty?
        hash
      end
    end

    ##
    # Returns the deserialized version of field_options
    #
    # @return [Array]
    #
    def options
      field_options.nil? ? nil : deserialize('Array', field_options)
    end


    private

    ##
    # Checks to make sure all values in field_options are of the correct type
    #
    # @return [Nil]
    #
    def field_options_types
      return if field_options.nil?
      options.each { |v| (field_options_type_error; return) unless v.class.to_s == field_type }
    end

    ##
    # Checks to make sure all values in field_options are unique
    #
    # @return [Nil]
    #
    def field_options_unique
      return if field_options.nil?
      field_options_unique_error unless options.uniq.count == options.count
    end

    ##
    # Checks that the default value is of the correct type
    #
    # @return [Nil]
    #
    def default_type
      return if default.nil?
      default_type_error unless is_correct_type?(field_type, default)
    end

    ##
    # Adds error if there is an field_options_error
    #
    # @return [Nil]
    #
    def field_options_type_error
      errors.add(:field_options, 'Value in options is not the correct type')
    end

    ##
    # Adds error if values in field_options aren't unique
    #
    # @return [Nil]
    #
    def field_options_unique_error
      errors.add(:field_options, 'Values in field options must be unique')
    end

    ##
    # Adds error if the default passed is not the correct type
    #
    # @return [Nil]
    #
    def default_type_error
      errors.add(:default, 'Default value is not the correct type')
    end

  end
end
