module CustomFielder
  class Value < ActiveRecord::Base
    include TypeCheckingHelper
    include DeserializationHelper

    before_validation :set_default

    validate :check_type
    validate :check_allow_blank
    validate :check_field_options
    validate :check_fieldable_type

    belongs_to :field
    belongs_to :fieldable, polymorphic: true

    scope :values_with_field, -> (field) { where(field: field) }
    scope :values_with_class, -> (klass) { where(fieldable_type: klass) }
    
    ##
    # Because everything is stored as string with ActiveRecord
    # we use a alias method chain to rename the value method that
    # uses the ActiveRecord String value to raw_value so when
    # value is called it returns the deserialized value
    #
    def value; super; end
    alias_method :raw_value, :value
    def value; deserialize_value; end;
    
    ##
    # Allows for the creation of many values at one time
    # this will return an array where the first element is
    # a boolean whose value corresponds to whether or not there
    # were any errors, the second value is the array of values
    # that were created
    #
    # @param values [Array]
    # @return [Array<Boolean,Array<Value>>]
    #
    def self.create_many(values)
      errors, values = false, transaction { create!(values) }
      values.each { |v| errors = true if v.errors.any? }

      [errors, values]
    end

    ##
    # Allows for updating many value records at one time
    # this will return an array where the first element is
    # a boolean whose value corresponds to whether or not there
    # were any errors, the second value is the array of values
    # that were updated
    # 
    # @param values [Array]
    # @return [Array<Boolean,Array<Value>>]
    #
    def self.update_many(values)
      errors, values = false, transaction { values.map { |v| update(v[:id], v) } }
      values.each { |v| errors = true if v.errors.any? }

      [errors, values]
    end

    ##
    # Allows for the deletion of many values at one time
    # this will return an array where the first element is a
    # boolean whose value corresponds to whether or not there
    # were any errors, the second value is the array of values
    # that were deleted
    #
    # @param ids [Array]
    # @return [Array<Boolean,Array<Value>>]
    #
    def self.delete_many(ids)
      errors, values = false, transaction { where(id: ids).destroy_all }
      values.each { |v| errors = true if v.errors.any? }

      [errors, values]
    end

    private

    ##
    # Gets the field type from the field attributes
    #
    # @return [String]
    #
    def type
      @type ||= field.field_type
    end

    ##
    # Gets the fieldable type from the field
    #
    # @return [String]
    #
    def field_fieldable_type
      @field_fieldable_type ||= field.fieldable_type
    end

    ##
    # Gets the boolean value of allow blank from field attributes
    #
    # @return [Boolean]
    #
    def allow_blank
      @allow_blank ||= field.allow_blank
    end

    ##
    # Gets the array of allowed values from field attributes
    #
    # @return [Array]
    #
    def field_options
      @field_options ||= field.options
    end

    ##
    # Gets the default value from the field attributes
    #
    # @return [String]
    #
    def default
      @default ||= field.default
    end

    ##
    # Deserializes the field value that is stored as a string
    #
    # @return [Array|Integer|Float|Date|DateTime|Boolean|String]
    #
    def deserialize_value
      return nil unless raw_value
      @deserialized_value ||= deserialize(type, raw_value)
    end

    ##
    # Sets the value to the default if value is blank and allow blank is false
    #
    # @return [Nil]
    #
    def set_default
      return if allow_blank
      update_attribute(:value, default) if default? and raw_value_blank?
    end

    ##
    # @return [Nil] Adds error if invalid
    #
    def check_type
      return if raw_value_blank?
      value_mismatch_error unless is_correct_type?(type, raw_value)
    end

    ##
    # @return [Nil] Adds error if invalid
    #
    def check_allow_blank
      value_blank_error if !allow_blank and raw_value_blank?
    end

    ##
    # @return [Nil] Adds error if invalid
    #
    def check_field_options
      return unless field_options
      value_not_allowed_error unless field_options.include?(value)
    end

    ##
    # @return [Nil] Adds error if invalid
    #
    def check_fieldable_type
      value_fieldable_type_error unless field_fieldable_type == fieldable_type
    end

    ##
    # If a value is passed that does not match field type adds
    # a mismatch error
    #
    # @return [Nil]
    #
    def value_mismatch_error
      errors.add(:value, 'Value passed does not match field type')
    end

    ##
    # If a value is passed that is blank and allow_blank is false
    # adds a value blank error
    #
    # @return [Nil]
    #
    def value_blank_error
      errors.add(:value, 'Blank value not allowed')
    end

    ##
    # If a value is passed that is not included in allowed values
    # adds a value not allowed error
    #
    # @return [Nil]
    #
    def value_not_allowed_error
      errors.add(:value, 'Value passed not in options')
    end

    ##
    # If a value is trying to be assigned to a model that is different
    # than the one the field was made for adds value fieldable type error
    #
    # @return [Nil]
    #
    def value_fieldable_type_error
      errors.add(:value, "Can't assign value to a #{fieldable_type} when field is for #{field_fieldable_type}")
    end

    ##
    # Return true if raw_value is nil or an empty string
    #
    # @return [Boolean]
    #
    def raw_value_blank?
      raw_value == nil or raw_value == ''
    end

    ##
    # Return true if default is set and is not an empty string
    #
    # @return [Boolean]
    #
    def default?
      default and default != ''
    end

  end
end
