module CustomFielder
  module CustomFieldable
    extend ActiveSupport::Concern

    included do
      has_many :custom_fielder_fields,
        through: :custom_fielder_values,
        class_name: 'CustomFielder::Field',
        source: :field

      has_many :custom_fielder_values,
        as: :fieldable,
        dependent: :destroy,
        class_name: 'CustomFielder::Value'


      accepts_nested_attributes_for :custom_fielder_values, allow_destroy: true
      accepts_nested_attributes_for :custom_fielder_fields, allow_destroy: false

      alias_method :custom_fields,       :custom_fielder_fields
      alias_method :custom_field_values, :custom_fielder_values

      ##
      # Returns all the objects on the model that have the passed
      # value, defined on the field with the passed field_id
      #
      # @return [ActiveRecord::Relation]
      #
      def self.find_objects_with_custom_values_where(value:, field_id:)
        joins(:custom_fielder_values).
        where(custom_fielder_values: {
          value: value,
          field_id: field_id
        })
      end

      ##
      # Provides a list of columns to be included in an
      # object search. This is to be able to dynamically
      # create tables no matter the object
      #
      # @note this method should be overridden
      #
      # @return [Array]
      #
      def self.custom_fieldable_columns
        [:id]
      end

      ##
      # Provides a uniform means of accessing data no matter
      # the object, this can be overridden to suit your
      # needs
      #
      # @note this method can be overridden more specific
      # @note object selection
      #
      # @return [ActiveRecord::Relation]
      #
      def self.custom_fieldable_search
        all
      end

      ##
      # Gets all the avaiable fieldable objects for a model
      # that do not already have a value defined on them
      #
      # @param field [CustomFielder::Field] field to search for values on
      #
      # @return [Array<Hash>]
      #
      def self.custom_fieldable_objects(field)
        custom_fieldable_search.where.not(id: field.values.select(:fieldable_id))
      end

      ##
      # Returns a hash where the keys are the field names
      # and the values are the field values
      #
      # @return [Hash]
      #
      def get_custom_fields_with_values
        custom_field_values.inject(Hash.new) do |h, field_value|
          field_name = field_value.field.name
          h[field_name] = field_value.value; h
        end
      end

      ##
      # Provides a shorter method call to retrieve a custom fields
      # value given the field. Returns the value or nil if the value
      # does not exist
      #
      # @return [String|Nil]
      #
      def value_for_custom_field(field)
        value = self.custom_fielder_values.find_by(field_id: field.id)
        value.nil? ? nil : value.value
      end

    end
  end
end
